@@ -2,7 +2,7 @@
 #
 #  PowerDNS web api python client and interface (python-powerdns)
 #
-#  Copyright (C) 2018 Denis Pompilio (jawa) <dpompilio@vente-privee.com>
+#  Copyright (C) 2018 Denis Pompilio (jawa) <denis.pompilio@gmail.com>
 #
 #  This file is part of python-powerdns
 #
@@ -24,6 +24,9 @@ powerdns.interface - PowerDNS API interface
 import logging
 import os
 import json
+import time
+from typing import List
+
 from .exceptions import PDNSCanonicalError
 
 
@@ -62,10 +65,10 @@ class PDNSEndpoint(PDNSEndpointBase):
         super(PDNSEndpoint, self).__init__(api_client)
 
     def __repr__(self):
-        return 'PDNSEndpoint(%s)' % self.api_client
+        return 'PDNSEndpoint(%s)' % repr(self.api_client)
 
     def __str__(self):
-        return 'PDNSEndpoint:%s' % self.api_client.api_endpoint
+        return str(self.api_client)
 
     @property
     def servers(self):
@@ -113,6 +116,7 @@ class PDNSServer(PDNSEndpointBase):
     """
     def __init__(self, api_client, api_data):
         """Initialization method"""
+        self._api_client = api_client
         self._api_data = api_data
         self.sid = api_data['id']
         self.version = api_data['version']
@@ -122,10 +126,12 @@ class PDNSServer(PDNSEndpointBase):
         super(PDNSServer, self).__init__(api_client)
 
     def __repr__(self):
-        return self.__str__()
+        return 'PDNSServer(%s, %s)' % (
+            repr(self._api_client), repr(self._api_data)
+        )
 
     def __str__(self):
-        return 'PDNSServer:%s' % self.sid
+        return self.sid
 
     @property
     def config(self):
@@ -251,8 +257,8 @@ class PDNSServer(PDNSEndpointBase):
     # pylint: disable=inconsistent-return-statements
     # TODO: Full implementation of zones endpoint
     def create_zone(self, name, kind, nameservers, masters=None, servers=None,
-                    rrsets=None):
-        """Create a new zone
+                    rrsets=None, update=False):
+        """Create or update a (new) zone
 
         :param str name: Name of zone
         :param str kind: Type of zone
@@ -260,6 +266,7 @@ class PDNSServer(PDNSEndpointBase):
         :param list masters: Zone masters
         :param list servers: List of forwarded-to servers (recursor only)
         :param list rrsets: Resource records sets
+        :param bool update: If the zone need to be updated or created
         :return: Created zone as :class:`PDNSZone` instance or :obj:`None`
 
         .. seealso:: https://doc.powerdns.com/md/httpapi/api_spec/#url-apiv1serversserver95idzones
@@ -276,13 +283,22 @@ class PDNSServer(PDNSEndpointBase):
         if rrsets:
             zone_data['rrsets'] = rrsets
 
-        LOG.info("creation of zone: %s", name)
-        zone_data = self._post("%s/zones" % self.url, data=zone_data)
+        if update is True:
+            LOG.info("update of zone: %s", name)
+            get_zone = self.get_zone(name).details
+            zone_id = get_zone['id']
+            zone_data = self._patch("{}/zones/{}".format(self.url,
+                                                         zone_id),
+                                    data=zone_data)
+
+        else:
+            LOG.info("creation of zone: %s", name)
+            zone_data = self._post("%s/zones" % self.url, data=zone_data)
 
         if zone_data:
             # reset server object cache
             self._zones = None
-            LOG.info("zone %s successfully created", name)
+            LOG.info("zone %s successfully processed", name)
             return PDNSZone(self.api_client, self, zone_data)
 
     def delete_zone(self, name):
@@ -328,6 +344,7 @@ class PDNSZone(PDNSEndpointBase):
     """
     def __init__(self, api_client, server, api_data):
         """Initialization method"""
+        self._api_client = api_client
         self._api_data = api_data
         self.server = server
         self.name = api_data['name']
@@ -336,10 +353,12 @@ class PDNSZone(PDNSEndpointBase):
         super(PDNSZone, self).__init__(api_client)
 
     def __repr__(self):
-        return self.__str__()
+        return "PDNSZone(%s, %s, %s)" % (
+            repr(self._api_client), repr(self.server), repr(self._api_data)
+        )
 
     def __str__(self):
-        return 'PDNSZone:%s:%s' % (self.server.sid, self.name)
+        return self.name
 
     @property
     def details(self):
@@ -361,14 +380,19 @@ class PDNSZone(PDNSEndpointBase):
         """Get record data
 
         :param str name: Record name
-        :return: Record data as :class:`dict` or :obj:`None`
+        :return: Records data as :func:`list`
         """
+        records = []
         LOG.info("getting zone record: %s", name)
         for record in self.details['rrsets']:
             if name == record['name']:
                 LOG.info("record found: %s", name)
-                return record
-        LOG.info("record not found: %s", name)
+                records.append(record)
+
+        if not records:
+            LOG.info("record not found: %s", name)
+
+        return records
 
     def create_records(self, rrsets):
         """Create resource record sets
@@ -386,7 +410,7 @@ class PDNSZone(PDNSEndpointBase):
         self._details = None
         return self._patch(self.url, data={'rrsets': rrsets})
 
-    def delete_record(self, rrsets):
+    def delete_records(self, rrsets):
         """Delete resource record sets
 
         :param list rrsets: Resource record sets
@@ -428,20 +452,28 @@ class PDNSZone(PDNSEndpointBase):
                 json.dump(self.details, backup_fp)
         LOG.info("zone %s successfully saved", self.name)
 
+    def notify(self):
+        """Trigger notification for zone updates"""
+        LOG.info("notify of zone: %s", self.name)
+        return self._put(self.url + '/notify')
+
 
 # pylint: disable=line-too-long
 class RRSet(dict):
-    """Resource record data for PowerDNS API
+    """
+    Resource record data for PowerDNS API
 
     :param str changetype: API keyword DELETE or REPLACE
     :param str name: Record name
     :param str rtype: Record type
     :param list records: List of Str or Tuple(content_str, disabled_bool)
     :param int ttl: Record time to live
+    :param list comments: Comments for this RRSet
 
     .. seealso:: https://doc.powerdns.com/md/httpapi/api_spec/#url-apiv1serversserver95idzoneszone95id
     """
-    def __init__(self, name, rtype, records, ttl=3600, changetype='REPLACE'):
+    def __init__(self, name, rtype, records, ttl=3600, changetype='REPLACE',
+                 comments: List["Comment"] = None):
         """Initialization"""
         LOG.debug("new rrset object for %s", name)
         super(RRSet, self).__init__()
@@ -453,23 +485,39 @@ class RRSet(dict):
         self['records'] = []
         for record in records:
             disabled = False
-            if isinstance(record, tuple):
+            if isinstance(record, tuple) or isinstance(record, list):
                 disabled = record[1]
                 record = record[0]
             self['records'].append({'content': record, 'disabled': disabled})
-
-    def __str__(self):
-        return "(ttl=%d) %s  %s  %s)" % (self['ttl'], self['name'],
-                                         self['type'],
-                                         [r[0] for r in self.raw_records])
+        if comments is None:
+            self["comments"] = list()
+        else:
+            self["comments"] = comments
 
     def __repr__(self):
-        return "powerdns.RRSet(\"%s\", \"%s\", \"%s\", %d, \"%s\")" % (
-            self['name'],
-            self['type'],
-            self.raw_records,
-            self['ttl'],
-            self['changetype'])
+        return "RRSet(%s, %s, %s, %s, %s, %s)" % (
+            repr(self['name']),
+            repr(self['type']),
+            repr(self.raw_records),
+            repr(self['ttl']),
+            repr(self['changetype']),
+            repr(self['comments']),
+        )
+
+    def __str__(self):
+        records = []
+
+        for rr in self.raw_records:
+            if isinstance(rr, tuple) or isinstance(rr, list):
+                records += [rr[0]]
+            else:
+                records += [rr]
+
+        return "(ttl=%d) %s  %s  %s %s)" % (self['ttl'],
+                                            self['name'],
+                                            self['type'],
+                                            records,
+                                            self['comments'],)
 
     def ensure_canonical(self, zone):
         """Ensure every record names are canonical
@@ -495,3 +543,33 @@ class RRSet(dict):
                     LOG.debug("transforming %s with %s",
                               record['content'], zone)
                     record['content'] += ".%s" % zone
+
+
+class Comment(dict):
+    """
+    Comment data for PowerDNS API RRSets
+
+
+    :param str content: the content of the comment
+    :param str account: the account
+    :param int modified_at: Unix timestamp at which the comment was last
+                            modified. Will be set to the current timestamp if
+                            None.
+
+    .. seealso:: https://doc.powerdns.com/md/httpapi/api_spec/#zone95collection
+    """
+
+    def __init__(self, content: str, account: str = "", modified_at: int = None):
+        self["content"] = content
+        self["account"] = account
+        if modified_at is None:
+            self["modified_at"] = int(time.time())
+        else:
+            self["modified_at"] = modified_at
+
+    def __repr__(self):
+        return "Comment(%s, %s, %s)" % (
+            repr(self["content"]),
+            repr(self["account"]),
+            repr(self["modified_at"]),
+        )
